@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: color.c,v 1.98.2.3 2013/03/10 18:23:59 broeker Exp $"); }
+static char *RCSid() { return RCSid("$Id: color.c,v 1.114 2014/05/13 18:26:40 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - color.c */
@@ -30,14 +30,13 @@ static char *RCSid() { return RCSid("$Id: color.c,v 1.98.2.3 2013/03/10 18:23:59
 #include "plot.h"
 #include "graph3d.h"
 #include "pm3d.h"
-#include "graphics.h"
 #include "term_api.h"
 #include "util3d.h"
 #include "alloc.h"
 
 /* COLOUR MODES - GLOBAL VARIABLES */
 
-t_sm_palette sm_palette;  /* initialized in init_color() */
+t_sm_palette sm_palette;  /* initialized in plot.c on program entry */
 
 /* Copy of palette previously in use.
  * Exported so that change_term() can invalidate contents
@@ -48,16 +47,11 @@ static t_sm_palette prev_palette = {
 	(rgb_color *) 0, -1
     };
 
-
-#ifdef EXTENDED_COLOR_SPECS
-int supply_extended_color_specs = 0;
-#endif
-
 /* Internal prototype declarations: */
 
-static void draw_inside_color_smooth_box_postscript __PROTO((FILE * out));
-static void draw_inside_color_smooth_box_bitmap __PROTO((FILE * out));
-void cbtick_callback __PROTO((AXIS_INDEX axis, double place, char *text, 
+static void draw_inside_color_smooth_box_postscript __PROTO((void));
+static void draw_inside_color_smooth_box_bitmap __PROTO((void));
+void cbtick_callback __PROTO((AXIS_INDEX axis, double place, char *text, int ticlevel,
 			struct lp_style_type grid, struct ticmark *userlabels));
 
 
@@ -79,15 +73,13 @@ init_color()
   sm_palette.use_maxcolors = 0;
   sm_palette.colors = 0;
   sm_palette.color = NULL;
-  sm_palette.ps_allcF = 0;
+  sm_palette.ps_allcF = FALSE;
   sm_palette.gradient_num = 0;
   sm_palette.gradient = NULL;
   sm_palette.cmodel = C_MODEL_RGB;
   sm_palette.Afunc.at = sm_palette.Bfunc.at = sm_palette.Cfunc.at = NULL;
+  sm_palette.colorMode = SMPAL_COLOR_MODE_RGB;
   sm_palette.gamma = 1.5;
-
-  /* initialisation of smooth color box */
-  color_box = default_color_box;
 }
 
 
@@ -117,8 +109,9 @@ make_palette()
 	    term->make_palette(&sm_palette);
 	    prev_palette = sm_palette;
 	    FPRINTF((stderr,"make_palette: calling term->make_palette for term with ncolors == 0\n"));
-	} else
+	} else {
 	    FPRINTF((stderr,"make_palette: skipping duplicate palette for term with ncolors == 0\n"));
+	}
 	return 0;
     }
 
@@ -184,8 +177,6 @@ void
 set_color(double gray)
 {
     t_colorspec color;
-    if (!(term->set_color))
-	return;
     color.type = TC_FRAC;
     color.value = gray;
     color.lt = 0;
@@ -193,18 +184,27 @@ set_color(double gray)
 }
 
 void
-set_rgbcolor(int rgblt)
+set_rgbcolor_var(unsigned int rgbvalue)
 {
     t_colorspec color;
-    if (!(term->set_color))
-	return;
     color.type = TC_RGB;
-    color.lt = rgblt;
-    color.value = 0;
-    term->set_color(&color);
+    *(unsigned int *)(&color.lt) = rgbvalue;
+    color.value = -1;	/* -1 flags that this came from "rgb variable" */
+    apply_pm3dcolor(&color, term);
 }
 
-void ifilled_quadrangle(gpiPoint* icorners)
+void
+set_rgbcolor_const(unsigned int rgbvalue)
+{
+    t_colorspec color;
+    color.type = TC_RGB;
+    *(unsigned int *)(&color.lt) = rgbvalue;
+    color.value = 0;	/* 0 flags that this is a constant color */
+    apply_pm3dcolor(&color, term);
+}
+
+void
+ifilled_quadrangle(gpiPoint* icorners)
 {
     if (default_fillstyle.fillstyle == FS_EMPTY)
 	icorners->style = FS_OPAQUE;
@@ -212,7 +212,7 @@ void ifilled_quadrangle(gpiPoint* icorners)
 	icorners->style = style_from_fill(&default_fillstyle);
     term->filled_polygon(4, icorners);
 
-    if (pm3d.hidden3d_tag) {
+    if (pm3d.border.l_type != LT_NODRAW) {
 	int i;
 
 	/* It should be sufficient to set only the color, but for some */
@@ -253,25 +253,27 @@ filled_quadrangle(gpdPoint * corners)
     ifilled_quadrangle(icorners);
 }
 
-
+#ifdef PM3D_CONTOURS
 /*
    Makes mapping from real 3D coordinates, passed as coords array,
    to 2D terminal coordinates, then draws filled polygon
  */
 void
-filled_polygon_3dcoords(int points, struct coordinate GPHUGE * coords)
+filled_polygon_common(int points, struct coordinate GPHUGE * coords, TBOOLEAN fixed, double z)
 {
     int i;
     double x, y;
     gpiPoint *icorners;
     icorners = gp_alloc(points * sizeof(gpiPoint), "filled_polygon3d corners");
     for (i = 0; i < points; i++) {
-	map3d_xy_double(coords[i].x, coords[i].y, coords[i].z, &x, &y);
+	if (fixed)
+	    z = coords[i].z;
+	map3d_xy_double(coords[i].x, coords[i].y, z, &x, &y);
 	icorners[i].x = x;
 	icorners[i].y = y;
     }
 #ifdef EXTENDED_COLOR_SPECS
-    if (supply_extended_color_specs) {
+    if ((term->flags & TERM_EXTENDED_COLOR)) {
 	icorners[0].spec.gray = -1;	/* force solid color */
     }
 #endif
@@ -283,6 +285,11 @@ filled_polygon_3dcoords(int points, struct coordinate GPHUGE * coords)
     free(icorners);
 }
 
+void
+filled_polygon_3dcoords(int points, struct coordinate GPHUGE * coords)
+{
+    filled_polygon_common(points, coords, FALSE, 0.0);
+}
 
 /*
    Makes mapping from real 3D coordinates, passed as coords array, but at z coordinate
@@ -291,28 +298,11 @@ filled_polygon_3dcoords(int points, struct coordinate GPHUGE * coords)
 void
 filled_polygon_3dcoords_zfixed(int points, struct coordinate GPHUGE * coords, double z)
 {
-    int i;
-    double x, y;
-    gpiPoint *icorners;
-    icorners = gp_alloc(points * sizeof(gpiPoint), "filled_polygon_zfix corners");
-    for (i = 0; i < points; i++) {
-	map3d_xy_double(coords[i].x, coords[i].y, z, &x, &y);
-	icorners[i].x = x;
-	icorners[i].y = y;
-    }
-#ifdef EXTENDED_COLOR_SPECS
-    if (supply_extended_color_specs) {
-	icorners[0].spec.gray = -1;	/* force solid color */
-    }
-#endif
-    if (default_fillstyle.fillstyle == FS_EMPTY)
-	icorners->style = FS_OPAQUE;
-    else
-	icorners->style = style_from_fill(&default_fillstyle);
-    term->filled_polygon(points, icorners);
-    free(icorners);
+    filled_polygon_common(points, coords, TRUE, z);
 }
 
+#endif /* PM3D_CONTOURS */
+
 
 /*
    Draw colour smooth box
@@ -327,30 +317,30 @@ filled_polygon_3dcoords_zfixed(int points, struct coordinate GPHUGE * coords, do
    PS routine.
  */
 static void
-draw_inside_color_smooth_box_postscript(FILE * out)
+draw_inside_color_smooth_box_postscript()
 {
     int scale_x = (color_box.bounds.xright - color_box.bounds.xleft), scale_y = (color_box.bounds.ytop - color_box.bounds.ybot);
     fputs("stroke gsave\t%% draw gray scale smooth box\n"
-	  "maxcolors 0 gt {/imax maxcolors def} {/imax 1024 def} ifelse\n", out);
+	  "maxcolors 0 gt {/imax maxcolors def} {/imax 1024 def} ifelse\n", gppsfile);
 
     /* nb. of discrete steps (counted in the loop) */
-    fprintf(out, "%i %i translate %i %i scale 0 setlinewidth\n", color_box.bounds.xleft, color_box.bounds.ybot, scale_x, scale_y);
+    fprintf(gppsfile, "%i %i translate %i %i scale 0 setlinewidth\n", color_box.bounds.xleft, color_box.bounds.ybot, scale_x, scale_y);
     /* define left bottom corner and scale of the box so that all coordinates
        of the box are from [0,0] up to [1,1]. Further, this normalization
        makes it possible to pass y from [0,1] as parameter to setgray */
-    fprintf(out, "/ystep 1 imax div def /y0 0 def /ii 0 def\n");
+    fprintf(gppsfile, "/ystep 1 imax div def /y0 0 def /ii 0 def\n");
     /* local variables; y-step, current y position and counter ii;  */
     if (sm_palette.positive == SMPAL_NEGATIVE)	/* inverted gray for negative figure */
-	fputs("{ 0.99999 y0 sub g ", out); /* 1 > x > 1-1.0/1024 */
+	fputs("{ 0.99999 y0 sub g ", gppsfile); /* 1 > x > 1-1.0/1024 */
     else
-	fputs("{ y0 g ", out);
+	fputs("{ y0 g ", gppsfile);
     if (color_box.rotation == 'v')
-	fputs("0 y0 N 1 0 V 0 ystep V -1 0 f\n", out);
+	fputs("0 y0 N 1 0 V 0 ystep V -1 0 f\n", gppsfile);
     else
-	fputs("y0 0 N 0 1 V ystep 0 V 0 -1 f\n", out);
+	fputs("y0 0 N 0 1 V ystep 0 V 0 -1 f\n", gppsfile);
     fputs("/y0 y0 ystep add def /ii ii 1 add def\n"
 	  "ii imax ge {exit} if } loop\n"
-	  "grestore 0 setgray\n", out);
+	  "grestore 0 setgray\n", gppsfile);
 }
 
 
@@ -361,7 +351,7 @@ draw_inside_color_smooth_box_postscript(FILE * out)
    over all thin rectangles
  */
 static void
-draw_inside_color_smooth_box_bitmap(FILE * out)
+draw_inside_color_smooth_box_bitmap()
 {
     int steps = 128; /* I think that nobody can distinguish more colours drawn in the palette */
     int i, j, xy, xy2, xy_from, xy_to;
@@ -369,7 +359,6 @@ draw_inside_color_smooth_box_bitmap(FILE * out)
     double xy_step, gray, range;
     gpiPoint corners[4];
 
-    (void) out;			/* to avoid "unused parameter" warning */
     if (color_box.rotation == 'v') {
 	corners[0].x = corners[3].x = color_box.bounds.xleft;
 	corners[1].x = corners[2].x = color_box.bounds.xright;
@@ -425,7 +414,7 @@ draw_inside_color_smooth_box_bitmap(FILE * out)
 	    corners[1].x = corners[2].x = GPMIN(xy_to,xy2+1);
 	}
 #ifdef EXTENDED_COLOR_SPECS
-	if (supply_extended_color_specs)
+	if ((term->flags & TERM_EXTENDED_COLOR))
 	    corners[0].spec.gray = -1;	/* force solid color */
 #endif
 	/* print the rectangle with the given colour */
@@ -444,10 +433,11 @@ cbtick_callback(
     AXIS_INDEX axis,
     double place,
     char *text,
+    int ticlevel,
     struct lp_style_type grid, /* linetype or -2 for no grid */
     struct ticmark *userlabels)
 {
-    int len = (text ? CB_AXIS.ticscale : CB_AXIS.miniticscale)
+    int len = TIC_SCALE(ticlevel, COLOR_AXIS)
 	* (CB_AXIS.tic_in ? -1 : 1) * (term->h_tic);
     double cb_place = (place - CB_AXIS.min) / (CB_AXIS.max - CB_AXIS.min);
 	/* relative z position along the colorbox axis */
@@ -555,7 +545,6 @@ void
 draw_color_smooth_box(int plot_mode)
 {
     double tmp;
-    FILE *out = gppsfile;	/* either gpoutfile or PSLATEX_auxfile */
 
     if (color_box.where == SMCOLOR_BOX_NO)
 	return;
@@ -645,13 +634,11 @@ draw_color_smooth_box(int plot_mode)
 	color_box.bounds.ybot = tmp;
     }
 
-    /* Optimized version of the smooth colour box in postscript. Advantage:
-       only few lines of code is written into the output file.
-     */
+    /* The PostScript terminal has an Optimized version */
     if ((term->flags & TERM_IS_POSTSCRIPT) != 0)
-	draw_inside_color_smooth_box_postscript(out);
+	draw_inside_color_smooth_box_postscript();
     else
-	draw_inside_color_smooth_box_bitmap(out);
+	draw_inside_color_smooth_box_bitmap();
 
     if (color_box.border) {
 	/* now make boundary around the colour box */
@@ -731,3 +718,46 @@ draw_color_smooth_box(int plot_mode)
 
 }
 
+/*
+ * User-callable builtin color conversion 
+ */
+void
+f_hsv2rgb(union argument *arg)
+{
+    struct value h, s, v, result;
+    rgb_color color = {0., 0., 0.};
+
+    (void) arg;
+    (void) pop(&v);
+    (void) pop(&s);
+    (void) pop(&h);
+
+    if (h.type == INTGR)
+	color.r = h.v.int_val;
+    else if (h.type == CMPLX)
+	color.r = h.v.cmplx_val.real;
+    if (s.type == INTGR)
+	color.g = s.v.int_val;
+    else if (s.type == CMPLX)
+	color.g = s.v.cmplx_val.real;
+    if (v.type == INTGR)
+	color.b = v.v.int_val;
+    else if (v.type == CMPLX)
+	color.b = v.v.cmplx_val.real;
+
+    if (color.r < 0)
+	color.r = 0;
+    if (color.g < 0)
+	color.g = 0;
+    if (color.b < 0)
+	color.b = 0;
+    if (color.r > 1.)
+	color.r = 1.;
+    if (color.g > 1.)
+	color.g = 1.;
+    if (color.b > 1.)
+	color.b = 1.;
+
+    (void) Ginteger(&result, hsv2rgb(&color));
+    push(&result);
+}
