@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: time.c,v 1.24.2.3 2014/01/27 21:54:07 sfeam Exp $"); }
+static char *RCSid() { return RCSid("$Id: time.c,v 1.28.2.1 2014/11/08 04:52:26 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - time.c */
@@ -52,67 +52,8 @@ static char *RCSid() { return RCSid("$Id: time.c,v 1.24.2.3 2014/01/27 21:54:07
 #include "util.h"
 #include "variable.h"
 
-/* build as a standalone test */
-
-#ifdef TEST_TIME
-
-# ifdef HAVE_SYS_TIMEB_H
-#  include <sys/timeb.h>
-# else
-/* declare struct timeb */
-extern int ftime(struct timeb *);
-# endif				/* !HAVE_SYS_TIMEB_H */
-
-# define int_error(x,y) fprintf(stderr, "Error: " y "\n")
-# define int_warn(x,y) fprintf(stderr, "Warn: " y "\n")
-
-/* need (only) these from plot.h */
-# define ZERO_YEAR	2000
-/* 1st jan, 2000 is a Saturday (cal 1 2000 on unix) */
-# define JAN_FIRST_WDAY 6
-
-/*  zero gnuplot (2000) - zero system (1970) */
-# define SEC_OFFS_SYS	946684800.0
-
-/* avg, incl. leap year */
-# define YEAR_SEC	31557600.0
-
-/* YEAR_SEC / 12 */
-# define MON_SEC		2629800.0
-
-# define WEEK_SEC	604800.0
-# define DAY_SEC		86400.0
-
-/* HBB 990826: moved definitions up here, to avoid 'extern' where
- * it is neither wanted nor needed */
-char *abbrev_month_names[] =
-{ "jan", "feb", "mar", "apr", "may", "jun", "jul",
-  "aug", "sep", "oct", "nov", "dec"
-};
-
-char *full_month_names[] =
-{ "January", "February", "March", "April", "May",
-  "June", "July", "August", "September", "October",
-  "November", "December"
-};
-
-char *abbrev_day_names[] =
-{ "sun", "mon", "tue", "wed", "thu", "fri", "sat"};
-
-char *full_day_names[] =
-{ "Sunday", "Monday", "Tuesday", "Wednesday",
-  "Thursday", "Friday", "Saturday"
-};
-
-#else /* TEST_TIME */
-
-/*  # include "setshow.h" */		/* for month names etc */
-
-#endif /* TEST_TIME */
-
 static char *read_int __PROTO((char *s, int nr, int *d));
 
-
 static char *
 read_int(char *s, int nr, int *d)
 {
@@ -137,7 +78,7 @@ static int gdysize __PROTO((int yr));
 
 static int mndday[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
 
-static size_t xstrftime __PROTO((char *buf, size_t bufsz, const char *fmt, struct tm * tm, double usec));
+static size_t xstrftime __PROTO((char *buf, size_t bufsz, const char *fmt, struct tm * tm, double usec, double fulltime));
 
 /* days in year */
 static int
@@ -386,7 +327,7 @@ gstrftime(char *s, size_t bsz, const char *fmt, double l_clock)
 
     usec = l_clock - (double)floor(l_clock);
 
-    return xstrftime(s, bsz, fmt, &tm, usec);
+    return xstrftime(s, bsz, fmt, &tm, usec, l_clock);
 }
 
 
@@ -396,18 +337,20 @@ xstrftime(
     size_t bsz,			/* space available */
     const char *fmt,
     struct tm *tm,
-    double usec)
+    double usec,
+    double fulltime)
 {
     size_t l = 0;			/* chars written so far */
-
+    int incr = 0;			/* chars just written */
     char *s = str;
+    TBOOLEAN sign_printed = FALSE;
 
-    memset(s, '\0', bsz);
+    memset(str, '\0', bsz);
 
     while (*fmt != '\0') {
 	if (*fmt != '%') {
-	    if (l >= bsz)
-		return (0);
+	    if (l >= bsz-1)
+		return 0;
 	    *s++ = *fmt++;
 	    l++;
 	} else {
@@ -459,8 +402,8 @@ xstrftime(
 			if (!z)					\
 			    z=(dz);				\
 		    }						\
-		    CHECK_SPACE(w);				\
-		    sprintf(s, z ? "%0*d" : "%*d", w, (x));	\
+		    incr = snprintf(s, bsz-l-1, z ? "%0*d" : "%*d", w, (x));	\
+		    CHECK_SPACE(incr);				\
 		} while(0)
 
 	    case '%':
@@ -485,27 +428,18 @@ xstrftime(
 		COPY_STRING(full_month_names[tm->tm_mon]);
 		break;
 
-
-#if 0
-		/* %x not currently supported, so neither is c */
-	    case 'c':
-		if (!xstrftime(s, bsz - l, "%x %X", tm, 0.))
-		    return (0);
-		break;
-#endif
-
 	    case 'd':
 		FORMAT_STRING(1, 2, tm->tm_mday);	/* %02d */
 		break;
 
 	    case 'D':
-		if (!xstrftime(s, bsz - l, "%m/%d/%y", tm, 0.))
-		    return (0);
+		if (!xstrftime(s, bsz - l, "%m/%d/%y", tm, 0., fulltime))
+		    return 0;
 		break;
 
 	    case 'F':
-		if (!xstrftime(s, bsz - l, "%Y-%m-%d", tm, 0.))
-		    return (0);
+		if (!xstrftime(s, bsz - l, "%Y-%m-%d", tm, 0., fulltime))
+		    return 0;
 		break;
 
 	    case 'H':
@@ -543,13 +477,13 @@ xstrftime(
 		break;
 
 	    case 'r':
-		if (!xstrftime(s, bsz - l, "%I:%M:%S %p", tm, 0.))
-		    return (0);
+		if (!xstrftime(s, bsz - l, "%I:%M:%S %p", tm, 0., fulltime))
+		    return 0;
 		break;
 
 	    case 'R':
-		if (!xstrftime(s, bsz - l, "%H:%M", tm, 0.))
-		    return (0);
+		if (!xstrftime(s, bsz - l, "%H:%M", tm, 0., fulltime))
+		    return 0;
 		break;
 
 	    case 's':
@@ -571,10 +505,77 @@ xstrftime(
 		break;
 
 	    case 'T':
-		if (!xstrftime(s, bsz - l, "%H:%M:%S", tm, 0.))
-		    return (0);
+		if (!xstrftime(s, bsz - l, "%H:%M:%S", tm, 0., fulltime))
+		    return 0;
 		break;
 
+	    case 't':		/* Time (as opposed to Date) formats */
+		{
+		int tminute, tsecond;
+
+		    switch (*fmt++) {
+		    case 'H':
+			/* +/- fractional hours (not wrapped at 24h) */
+			if (p > 0) {
+			    incr = snprintf(s, bsz-l-1, "%*.*f", w, p, fulltime/3600.);
+			    CHECK_SPACE(incr);
+			    break;
+			}
+			/* Set flag in case minutes come next */
+			if (fulltime < 0) {
+			    CHECK_SPACE(1);	/* the minus sign */
+			    sign_printed = TRUE;
+			    *s++ = '-';
+			    l++;
+			}
+			/* +/- integral hour truncated toward zero */
+			sprintf(s, "%0*d", w, (int)floor(fabs(fulltime/3600.)));
+			break;
+		    case 'M':
+			/* +/- fractional minutes (not wrapped at 60m) */
+			if (p > 0) {
+			    incr = snprintf(s, bsz-l-1, "%*.*f", w, p,
+				    sign_printed ? fabs(fulltime)/60. : fulltime/60.);
+			    CHECK_SPACE(incr);
+			    break;
+			}
+			/* +/- integral minute truncated toward zero */
+			tminute = floor(60. * (fabs(fulltime/3600.) - floor(fabs(fulltime/3600.))));
+			if (fulltime < 0) {
+			    if (!sign_printed) {
+				sign_printed = TRUE;
+				*s++ = '-';
+				l++;
+			    }
+			}
+			FORMAT_STRING(1, 2, tminute);	/* %02d */
+			break;
+		    case 'S':
+			/* +/- fractional seconds */
+			tsecond = floor(60. * (fabs(fulltime/60.) - floor(fabs(fulltime/60.))));
+			if (fulltime < 0) {
+			    if (usec > 0)
+				usec = 1.0 - usec;
+			    if (!sign_printed) {
+				*s++ = '-';
+				l++;
+			    }
+			}
+			FORMAT_STRING(1, 2, tsecond);	/* %02d */
+			if (p > 0) {
+			    double base = pow(10., (double)p);
+			    int msec = floor(0.5 + base * usec);
+			    char *f = &s[strlen(s)];
+			    CHECK_SPACE(p+1);
+			    sprintf(f, ".%0*d", p, msec<(int)base?msec:(int)base-1);
+			}
+			break;
+		    default:
+			break;
+		    }
+		    break;
+		}
+
 	    case 'W':		/* mon 1 day of week */
 		{
 		    int week;
@@ -639,11 +640,6 @@ xstrftime(
 		FORMAT_STRING(1, 4, tm->tm_year);	/* %04d */
 		break;
 
-#if 0
-	    case 'Z':
-		COPY_STRING(tm->tm_zone);
-		break;
-#endif
 	    }			/* switch */
 
 	    while (*s != '\0') {
@@ -665,7 +661,7 @@ double
 gtimegm(struct tm *tm)
 {
     int i;
-    /* returns sec from year ZERO_YEAR, defined in plot.h */
+    /* returns sec from year ZERO_YEAR, defined in gp_time.h */
     double dsec = 0.;
 
     if (tm->tm_year < ZERO_YEAR) {
@@ -860,37 +856,3 @@ gstrptime(char *s, char *fmt, struct tm *tm)
 
 
 #endif /* USE_SYSTEM_TIME */
-
-
-#ifdef TEST_TIME
-
-/* either print current time using supplied format, or read
- * supplied time using supplied format
- */
-
-
-int
-main(int argc, char *argv[])
-{
-    char output[80];
-
-    if (argc < 2) {
-	fputs("usage : test 'format' ['time']\n", stderr);
-	exit(EXIT_FAILURE);
-    }
-    if (argc == 2) {
-	struct timeb now;
-	struct tm *tm;
-	ftime(&now);
-	tm = gmtime(&now.time);
-	xstrftime(output, 80, argv[1], tm, 0.);
-	puts(output);
-    } else {
-	struct tm tm;
-	gstrptime(argv[2], argv[1], &tm);
-	puts(asctime(&tm));
-    }
-    exit(EXIT_SUCCESS);
-}
-
-#endif /* TEST_TIME */
