@@ -1,5 +1,5 @@
 /*
- * $Id: gp_cairo.c,v 1.60.2.6 2013/12/13 05:44:09 sfeam Exp $
+ * $Id: gp_cairo.c,v 1.87.2.2 2014/12/13 23:42:53 sfeam Exp $
  */
 
 /* GNUPLOT - gp_cairo.c */
@@ -133,7 +133,7 @@ static rgb_color gp_cairo_colorlist[12] = {
 {0,0,1}, /* blue */
 {1,0,1}, /* magenta */
 {0,1,1}, /* cyan */
-{1,1,0}, /* yellow */ 
+{1,1,0}, /* yellow */
 {0,0,0}, /* black */
 {1,0.3,0}, /* orange */
 {0.5,0.5,0.5} /* grey */
@@ -159,11 +159,11 @@ void gp_cairo_initialize_plot(plot_struct *plot)
 	plot->xscale = 1.0; plot->yscale = 1.0;
 	plot->device_xmax = 1; plot->device_ymax = 1;
 	plot->xmax = 1; plot->ymax = 1;
-	
+
 	plot->justify_mode = LEFT;
 	plot->linetype = 1;
 	plot->linewidth = 1.0;
-	plot->linestyle = 1;
+	plot->linestyle = GP_CAIRO_SOLID;
 	plot->pointsize = 1.0;
 	plot->dashlength = 1.0;
 	plot->text_angle = 0.0;
@@ -172,6 +172,8 @@ void gp_cairo_initialize_plot(plot_struct *plot)
 
 	plot->opened_path = FALSE;
 
+	plot->current_x = -1; plot->current_y = -1;
+
 	strncpy(plot->fontname, "", sizeof(plot->fontname));
 	plot->fontsize = 1.0;
 	plot->encoding = S_ENC_DEFAULT;
@@ -183,7 +185,7 @@ void gp_cairo_initialize_plot(plot_struct *plot)
 	plot->oversampling = TRUE;
 	plot->oversampling_scale = GP_CAIRO_SCALE;
 
-	plot->rounded = FALSE;
+	plot->linecap = BUTT;
 
 	plot->hinting = 100;
 
@@ -196,6 +198,12 @@ void gp_cairo_initialize_plot(plot_struct *plot)
 	plot->interrupt = FALSE;
 }
 
+#ifdef EAM_BOXED_TEXT
+/* Boxed text support */
+static int bounding_box[4];
+static double bounding_xmargin = 1.0;
+static double bounding_ymargin = 1.0;
+#endif
 
 /* set the transformation matrix of the context, and other details */
 /* NOTE : depends on the setting of xscale and yscale */
@@ -221,26 +229,33 @@ void gp_cairo_initialize_context(plot_struct *plot)
 	cairo_set_matrix(plot->cr, &matrix);
 
 	/* Default is square caps, mitered joins */
-	if (plot->rounded) {
+	if (plot->linecap == ROUNDED) {
 	    cairo_set_line_cap  (plot->cr, CAIRO_LINE_CAP_ROUND);
 	    cairo_set_line_join (plot->cr, CAIRO_LINE_JOIN_ROUND);
-	} else {
+	} else if (plot->linecap == SQUARE) {
 	    cairo_set_line_cap  (plot->cr, CAIRO_LINE_CAP_SQUARE);
 	    cairo_set_line_join (plot->cr, CAIRO_LINE_JOIN_MITER);
 	    cairo_set_miter_limit(plot->cr, 3.8);
+	} else {
+	    cairo_set_line_cap  (plot->cr, CAIRO_LINE_CAP_BUTT);
+	    cairo_set_line_join (plot->cr, CAIRO_LINE_JOIN_MITER);
+	    cairo_set_miter_limit(plot->cr, 3.8);
 	}
 
 }
 
 
-void gp_cairo_set_color(plot_struct *plot, rgb_color color)
+void gp_cairo_set_color(plot_struct *plot, rgb_color color, double alpha)
 {
 	/*stroke any open path */
 	gp_cairo_stroke(plot);
 
 	FPRINTF((stderr,"set_color %lf %lf %lf\n",color.r, color.g, color.b));
 
-	plot->color = color;
+	plot->color.r = color.r;
+	plot->color.g = color.g;
+	plot->color.b = color.b;
+	plot->color.alpha = alpha;
 }
 
 
@@ -291,7 +306,7 @@ void gp_cairo_set_font(plot_struct *plot, const char *name, int fontsize)
     char *c;
     char *fname;
 
-	FPRINTF(("set_font \"%s\" %d\n", name,fontsize));
+	FPRINTF((stderr,"set_font \"%s\" %d\n", name,fontsize));
 
 	/* Split out Bold and Italic attributes from font name */
 	fname = strdup(name);
@@ -301,18 +316,19 @@ void gp_cairo_set_font(plot_struct *plot, const char *name, int fontsize)
 		do { *d = *(d+1); } while (*d++);
 	    } else {
 		if (*c == '-') *c = ' ';
+		if (*c == ':') *c = ' ';
 	    }
 	}
 	if ((c = strstr(fname, " Bold"))) {
 	    do { *c = *(c+5); } while (*c++);
 	    plot->fontweight = PANGO_WEIGHT_BOLD;
 	} else
-	    plot->fontweight = 0;
+	    plot->fontweight = PANGO_WEIGHT_NORMAL;
 	if ((c = strstr(fname, " Italic"))) {
 	    do { *c = *(c+7); } while (*c++);
 	    plot->fontstyle = PANGO_STYLE_ITALIC;
 	} else
-	    plot->fontstyle = 0;
+	    plot->fontstyle = PANGO_STYLE_NORMAL;
 
 	strncpy( plot->fontname, fname, sizeof(plot->fontname) );
 	plot->fontsize = fontsize;
@@ -322,14 +338,20 @@ void gp_cairo_set_font(plot_struct *plot, const char *name, int fontsize)
 
 void gp_cairo_set_linewidth(plot_struct *plot, double linewidth)
 {
+	FPRINTF((stderr,"set_linewidth %lf\n",linewidth));
+
 	/*stroke any open path */
 	gp_cairo_stroke(plot);
 	/* draw any open polygon set */
 	gp_cairo_end_polygon(plot);
 
-	FPRINTF((stderr,"set_linewidth %lf\n",linewidth));
-
+	if (linewidth < 0.25)	/* Admittedly arbitrary */
+	    linewidth = 0.25;
 	plot->linewidth = linewidth;
+
+	if (!strcmp(term->name,"pdfcairo"))
+		plot->linewidth *= 2;
+
 }
 
 
@@ -366,20 +388,20 @@ void gp_cairo_draw_polygon(plot_struct *plot, int n, gpiPoint *corners)
 {
 	/* begin by stroking any open path */
 	gp_cairo_stroke(plot);
-	
+
 	if (plot->polygons_saturate) {
 		int i;
 		path_item *path;
-	
+
 		path = (path_item*) gp_alloc(sizeof(path_item), "gp_cairo : polygon path");
-	
+
 		path->n = n;
 		path->corners = (gpiPoint*) gp_alloc(n*sizeof(gpiPoint), "gp_cairo : polygon corners");
 		for(i=0;i<n;i++)
 			*(path->corners + i) = *corners++;
-	
+
 		path->color = plot->color;
-	
+
 		if (plot->polygon_path_last == NULL) {
 			FPRINTF((stderr,"creating a polygon path\n"));
 			path->previous = NULL;
@@ -408,7 +430,7 @@ void gp_cairo_end_polygon(plot_struct *plot)
 	int i;
 	path_item *path;
 	path_item *path2;
-	rgb_color color_sav;
+	rgba_color color_sav;
 	cairo_t *context;
 	cairo_t *context_sav;
 	cairo_surface_t *surface;
@@ -516,9 +538,8 @@ void gp_cairo_end_polygon(plot_struct *plot)
 	cairo_paint( plot->cr );
 }
 
-void gp_cairo_stroke(plot_struct *plot)
+void gp_cairo_set_dashtype(plot_struct *plot, int type, t_dashtype *custom_dash_type)
 {
-	double dashes[8];
 	static double dashpattern[4][8] =
 	{
 	    {5, 8, 5, 8, 5, 8, 5, 8},	/* Medium dash */
@@ -526,6 +547,58 @@ void gp_cairo_stroke(plot_struct *plot)
 	    {8, 4, 2, 4, 8, 4, 2, 4},	/* dash dot */
 	    {9, 4, 1, 4, 1, 4, 0, 0}	/* dash dot dot */
 	};
+	int lt = (type) % 5;
+
+	FPRINTF((stderr,"set_dashtype %d\n", type));
+
+	if (type == DASHTYPE_CUSTOM && custom_dash_type) {
+		/* Convert to internal representation */
+		int i;
+		double empirical_scale;
+
+		if (!strcmp(term->name,"pngcairo"))
+			empirical_scale = 0.25;
+		else
+			empirical_scale = 0.55;
+
+		if (plot->linewidth > 1)
+			empirical_scale *= plot->linewidth;
+
+		for (i=0; i<8; i++)
+			plot->current_dashpattern[i] = custom_dash_type->pattern[i]
+				* plot->dashlength
+				* plot->oversampling_scale * empirical_scale;
+		FPRINTF((stderr,"gp_cairo_set_dashtype: custom pattern\n"));
+		FPRINTF((stderr,"	%f %f %f %f %f %f %f %f\n",
+			plot->current_dashpattern[0], plot->current_dashpattern[1],
+			plot->current_dashpattern[2], plot->current_dashpattern[3],
+			plot->current_dashpattern[4], plot->current_dashpattern[5],
+			plot->current_dashpattern[6], plot->current_dashpattern[7]));
+		plot->linestyle = GP_CAIRO_DASH;
+
+	} else if (type > 0 && lt != 0) {
+		/* Use old (version 4) set of linetype patterns */
+		int i;
+		double empirical_scale = 1.;
+		if (plot->linewidth > 1)
+			empirical_scale *= plot->linewidth;
+
+		for (i=0; i<8; i++)
+			plot->current_dashpattern[i] = dashpattern[lt-1][i]
+				* plot->dashlength
+				* plot->oversampling_scale
+				* empirical_scale;
+		plot->linestyle = GP_CAIRO_DASH;
+
+	} else {
+		/* Every 5th pattern in the old set is solid */
+		plot->linestyle = GP_CAIRO_SOLID;
+	}
+
+}
+
+void gp_cairo_stroke(plot_struct *plot)
+{
 	int lt = plot->linetype;
 	double lw = plot->linewidth * plot->oversampling_scale;
 
@@ -545,23 +618,21 @@ void gp_cairo_stroke(plot_struct *plot)
 
 	else if (lt == LT_AXIS || plot->linestyle == GP_CAIRO_DOTS) {
 		/* Grid lines (lt 0) */
-		dashes[0] = 0.4 * plot->oversampling_scale * plot->dashlength;
-		dashes[1] = 4.0 * plot->oversampling_scale * plot->dashlength;
+		double dashes[2];
+		double empirical_scale = 1.0;
+		if (plot->linewidth > 1)
+			empirical_scale *= plot->linewidth;
+		dashes[0] = 0.4 * plot->oversampling_scale * plot->dashlength * empirical_scale;
+		dashes[1] = 4.0 * plot->oversampling_scale * plot->dashlength * empirical_scale;
 		cairo_set_dash(plot->cr, dashes, 2 /*num_dashes*/, 0 /*offset*/);
-		lw *= 0.6;
 	}
 
-	else if (plot->linestyle == GP_CAIRO_DASH
-		&& lt >= 0 && (lt % 5 != 0)) {
-		int i;
-		for (i=0; i<8; i++)
-			dashes[i] = dashpattern[(lt-1) % 5][i] 
-				* plot->dashlength
-				* plot->oversampling_scale;
-		cairo_set_dash(plot->cr, dashes, 8 /*num_dashes*/, 0 /*offset*/);
+	else if (plot->linestyle == GP_CAIRO_DASH) {
+		cairo_set_dash(plot->cr, &(plot->current_dashpattern[0]), 8 /*num_dashes*/, 0 /*offset*/);
 	}
 
-	cairo_set_source_rgb(plot->cr, plot->color.r, plot->color.g, plot->color.b);
+	cairo_set_source_rgba(plot->cr, plot->color.r, plot->color.g, plot->color.b,
+				1. - plot->color.alpha);
 	cairo_set_line_width(plot->cr, lw);
 
 	cairo_stroke(plot->cr);
@@ -574,13 +645,16 @@ void gp_cairo_stroke(plot_struct *plot)
 
 void gp_cairo_move(plot_struct *plot, int x, int y)
 {
+	/* Dec 2014 - Do not let zero-length moves interrupt     */
+	/* the current line/polyline context, e.g. dash pattern. */
+	if (x == plot->current_x && y == plot->current_y)
+		return;
+
 	/* begin by stroking any open path */
 	gp_cairo_stroke(plot);
 	/* also draw any open polygon set */
 	gp_cairo_end_polygon(plot);
 
-	FPRINTF((stderr,"move\n"));
-
 	plot->current_x = x;
 	plot->current_y = y;
 	plot->orig_current_x = x;
@@ -606,7 +680,7 @@ void gp_cairo_vector(plot_struct *plot, int x, int y)
 			fprintf(stderr,"wxt terminal : hinting error, setting to default\n");
 			plot->hinting = 100;
 		}
-		
+
 		/* detect and handle vertical lines */
 		/* the second test is there to avoid artefacts when you choose
 		* a high sampling ('set samples 10000'), so that a smooth function
@@ -653,7 +727,7 @@ void gp_cairo_vector(plot_struct *plot, int x, int y)
 /* pango needs a string encoded in utf-8. We use g_convert from glib.
  * gp_cairo_get_encoding() gives the encoding set via 'set enconding'
  * memory allocated for the string has to be freed */
-gchar * gp_cairo_convert(plot_struct *plot, const char* string)
+static gchar * gp_cairo_convert(plot_struct *plot, const char* string)
 {
 	gsize bytes_read;
 	GError *error = NULL;
@@ -701,19 +775,19 @@ gchar * gp_cairo_convert(plot_struct *plot, const char* string)
  * The following #ifdef WIN32 section is all to work around a bug in
  * the cairo/win32 backend for font rendering.  It has the effect of
  * testing for libfreetype support, and using that instead if possible.
- * Suggested by cairo developer Behdad Esfahbod. 
+ * Suggested by cairo developer Behdad Esfahbod.
  * Allin Cottrell suggests that this not necessary anymore for newer
  * versions of cairo.
  */
 #if defined(WIN32) && (CAIRO_VERSION_MAJOR < 2) && (CAIRO_VERSION_MINOR < 10)
-PangoLayout *
-gp_cairo_create_layout (cairo_t *cr)
+static PangoLayout *
+gp_cairo_create_layout(cairo_t *cr)
 {
     static PangoFontMap *fontmap;
     PangoContext *context;
     PangoLayout *layout;
 
-    if (fontmap == NULL) {   
+    if (fontmap == NULL) {
         fontmap = pango_cairo_font_map_new_for_font_type(CAIRO_FONT_TYPE_FT);
         if (fontmap == NULL) {
 	    fontmap = pango_cairo_font_map_get_default();
@@ -732,14 +806,15 @@ gp_cairo_create_layout (cairo_t *cr)
     return layout;
 }
 #else
-PangoLayout *
-gp_cairo_create_layout (cairo_t *cr)
+static PangoLayout *
+gp_cairo_create_layout(cairo_t *cr)
 {
     return pango_cairo_create_layout(cr);
 }
 #endif
 
-void gp_cairo_draw_text(plot_struct *plot, int x1, int y1, const char* string)
+void gp_cairo_draw_text(plot_struct *plot, int x1, int y1, const char* string,
+		    int *width, int *height)
 {
 	double x,y;
 	double arg = plot->text_angle * M_PI/180;
@@ -777,7 +852,7 @@ void gp_cairo_draw_text(plot_struct *plot, int x1, int y1, const char* string)
 
 	/* Create a PangoLayout, set the font and text */
 	layout = gp_cairo_create_layout (plot->cr);
-	
+
 	pango_layout_set_text (layout, string_utf8, -1);
 	g_free(string_utf8);
 	desc = pango_font_description_new ();
@@ -789,14 +864,18 @@ void gp_cairo_draw_text(plot_struct *plot, int x1, int y1, const char* string)
 #endif /*MAP_SYMBOL*/
 	pango_font_description_set_size (desc, (int) (plot->fontsize*PANGO_SCALE*plot->oversampling_scale) );
 
-	pango_font_description_set_weight (desc, 
-		plot->fontweight ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL);
-	pango_font_description_set_style (desc, 
+	pango_font_description_set_weight (desc, plot->fontweight);
+	pango_font_description_set_style (desc,
 		plot->fontstyle ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL);
 	pango_layout_set_font_description (layout, desc);
+	FPRINTF((stderr, "pango font description: %s\n", pango_font_description_to_string(desc)));
 	pango_font_description_free (desc);
-	
+
 	pango_layout_get_extents(layout, &ink_rect, &logical_rect);
+	if (width)
+		*width = logical_rect.width / PANGO_SCALE;
+	if (height)
+		*height = logical_rect.height / PANGO_SCALE;
 
 	/* EAM Mar 2009 - Adjusting the vertical position for every character fragment	*/
 	/* leads to uneven baselines. Better to adjust to the "average" character height */
@@ -836,7 +915,8 @@ void gp_cairo_draw_text(plot_struct *plot, int x1, int y1, const char* string)
 #endif /* helper point */
 
 	cairo_save (plot->cr);
-	cairo_set_source_rgb(plot->cr, plot->color.r, plot->color.g, plot->color.b);
+	cairo_set_source_rgba(plot->cr, plot->color.r, plot->color.g, plot->color.b,
+				1. - plot->color.alpha);
 	cairo_move_to (plot->cr, x-0.5, y-0.5);
 	cairo_rotate(plot->cr, -arg);
 
@@ -847,7 +927,7 @@ void gp_cairo_draw_text(plot_struct *plot, int x1, int y1, const char* string)
 	 * Do it by ourselves, or we can get spurious lines on future calls. */
 	cairo_new_path(plot->cr);
 
-#if 0 /* helper boxes to understand how text is positionned */
+#ifdef EAM_BOXED_TEXT
 	cairo_set_line_width(plot->cr, plot->linewidth*plot->oversampling_scale);
 	cairo_rotate(plot->cr, arg);
 	cairo_translate(plot->cr, x, y);
@@ -855,26 +935,25 @@ void gp_cairo_draw_text(plot_struct *plot, int x1, int y1, const char* string)
 
 	{
 	PangoRectangle ink, logical;
-	double lw = cairo_get_line_width (plot->cr);
-	pango_layout_get_extents (layout, &ink, &logical);
+	pango_layout_get_pixel_extents (layout, &ink, &logical);
 
-	cairo_set_source_rgba (plot->cr, 1.0, 0.0, 0.0, 0.75);	
-	cairo_rectangle (plot->cr,
-			(double)logical.x / PANGO_SCALE - lw / 2,
-			(double)logical.y / PANGO_SCALE - lw / 2,
-			(double)logical.width / PANGO_SCALE + lw,
-			(double)logical.height / PANGO_SCALE + lw);
-	cairo_stroke (plot->cr);
-	
-	cairo_set_source_rgba (plot->cr, 0.0, 1.0, 0.0, 0.75);
-	cairo_rectangle (plot->cr,
-			(double)ink.x / PANGO_SCALE - lw / 2,
-			(double)ink.y / PANGO_SCALE - lw / 2,
-			(double)ink.width / PANGO_SCALE + lw,
-			(double)ink.height / PANGO_SCALE + lw);
-	cairo_stroke (plot->cr);
+	/* Auto-initialization */
+	if (bounding_box[0] < 0 && bounding_box[1] < 0) {
+	    bounding_box[0] = bounding_box[2] = x;
+	    bounding_box[1] = bounding_box[3] = y;
 	}
-#endif /* helper boxes to understand how text is positionned */
+
+	/* Would it look better to use logical bounds rather than ink? */
+	if (bounding_box[0] > x + ink.x)
+	    bounding_box[0] = x + ink.x;
+	if (bounding_box[2] < x + ink.x + ink.width)
+	    bounding_box[2] = x + ink.x + ink.width;
+	if (bounding_box[1] > y + ink.y)
+	    bounding_box[1] = y + ink.y;
+	if (bounding_box[3] < y + ink.y + ink.height)
+	    bounding_box[3] = y + ink.y + ink.height;
+	}
+#endif
 
 	/* free the layout object */
 	g_object_unref (layout);
@@ -916,7 +995,8 @@ void gp_cairo_draw_point(plot_struct *plot, int x1, int y1, int style)
 
 	cairo_save(plot->cr);
 	cairo_set_line_width(plot->cr, plot->linewidth*plot->oversampling_scale);
-	cairo_set_source_rgb(plot->cr, plot->color.r, plot->color.g, plot->color.b);
+	cairo_set_source_rgba(plot->cr, plot->color.r, plot->color.g, plot->color.b,
+				1. - plot->color.alpha);
 
 	/* Dot	FIXME: because this is drawn as a filled circle, it's quite slow */
 	if (style < 0) {
@@ -924,7 +1004,9 @@ void gp_cairo_draw_point(plot_struct *plot, int x1, int y1, int style)
 		cairo_fill (plot->cr);
 	}
 
-	switch (style%13) {
+
+	style = style % 15;
+	switch (style) {
 	case 0: /* plus */
 		cairo_move_to(plot->cr, x-size, y);
 		cairo_line_to(plot->cr, x+size,y);
@@ -956,20 +1038,14 @@ void gp_cairo_draw_point(plot_struct *plot, int x1, int y1, int style)
 		cairo_stroke(plot->cr);
 		break;
 	case 3: /* box */
-		cairo_move_to(plot->cr, x-size, y-size);
-		cairo_line_to(plot->cr, x-size,y+size);
-		cairo_line_to(plot->cr, x+size,y+size);
-		cairo_line_to(plot->cr, x+size,y-size);
-		cairo_close_path(plot->cr);
-		cairo_stroke(plot->cr);
-		break;
 	case 4: /* filled box */
 		cairo_move_to(plot->cr, x-size, y-size);
 		cairo_line_to(plot->cr, x-size,y+size);
 		cairo_line_to(plot->cr, x+size,y+size);
 		cairo_line_to(plot->cr, x+size,y-size);
 		cairo_close_path(plot->cr);
-		cairo_fill_preserve(plot->cr);
+		if (style == 4)
+			cairo_fill_preserve(plot->cr);
 		cairo_stroke(plot->cr);
 		break;
 	case 5: /* circle */
@@ -982,52 +1058,48 @@ void gp_cairo_draw_point(plot_struct *plot, int x1, int y1, int style)
 		cairo_stroke(plot->cr);
 		break;
 	case 7: /* triangle */
-		cairo_move_to(plot->cr, x-size, y+size-plot->oversampling_scale);
-		cairo_line_to(plot->cr, x,y-size);
-		cairo_line_to(plot->cr, x+size,y+size-plot->oversampling_scale);
-		cairo_close_path(plot->cr);
-		cairo_stroke(plot->cr);
-		break;
 	case 8: /* filled triangle */
 		cairo_move_to(plot->cr, x-size, y+size-plot->oversampling_scale);
 		cairo_line_to(plot->cr, x,y-size);
 		cairo_line_to(plot->cr, x+size,y+size-plot->oversampling_scale);
 		cairo_close_path(plot->cr);
-		cairo_fill_preserve(plot->cr);
+		if (style == 8)
+			cairo_fill_preserve(plot->cr);
 		cairo_stroke(plot->cr);
 		break;
 	case 9: /* upside down triangle */
-		cairo_move_to(plot->cr, x-size, y-size+plot->oversampling_scale);
-		cairo_line_to(plot->cr, x,y+size);
-		cairo_line_to(plot->cr, x+size,y-size+plot->oversampling_scale);
-		cairo_close_path(plot->cr);
-		cairo_stroke(plot->cr);
-		break;
 	case 10: /* filled upside down triangle */
 		cairo_move_to(plot->cr, x-size, y-size+plot->oversampling_scale);
 		cairo_line_to(plot->cr, x,y+size);
 		cairo_line_to(plot->cr, x+size,y-size+plot->oversampling_scale);
 		cairo_close_path(plot->cr);
-		cairo_fill_preserve(plot->cr);
+		if (style == 10)
+			cairo_fill_preserve(plot->cr);
 		cairo_stroke(plot->cr);
 		break;
 	case 11: /* diamond */
-		cairo_move_to(plot->cr, x-size, y);
-		cairo_line_to(plot->cr, x,y+size);
-		cairo_line_to(plot->cr, x+size,y);
-		cairo_line_to(plot->cr, x,y-size);
-		cairo_close_path(plot->cr);
-		cairo_stroke(plot->cr);
-		break;
 	case 12: /* filled diamond */
 		cairo_move_to(plot->cr, x-size, y);
 		cairo_line_to(plot->cr, x,y+size);
 		cairo_line_to(plot->cr, x+size,y);
 		cairo_line_to(plot->cr, x,y-size);
 		cairo_close_path(plot->cr);
-		cairo_fill_preserve(plot->cr);
+		if (style == 12)
+			cairo_fill_preserve(plot->cr);
 		cairo_stroke(plot->cr);
 		break;
+	case 13: /* pentagon */
+	case 14: /* filled pentagon */
+		cairo_move_to(plot->cr, x+size*0.5878, y-size*0.8090);
+		cairo_line_to(plot->cr, x-size*0.5878, y-size*0.8090);
+		cairo_line_to(plot->cr, x-size*0.9511, y+size*0.3090);
+		cairo_line_to(plot->cr, x,             y+size);
+		cairo_line_to(plot->cr, x+size*0.9511, y+size*0.3090);
+		cairo_close_path(plot->cr);
+		if (style == 14)
+			cairo_fill_preserve(plot->cr);
+		cairo_stroke(plot->cr);
+		break;				
 	default :
 		break;
 	}
@@ -1124,6 +1196,7 @@ void gp_cairo_draw_image(plot_struct *plot, unsigned int * image, int x1, int y1
 void gp_cairo_add_attr(plot_struct *plot, PangoAttrList * AttrList, int start, int end )
 {
 	PangoAttribute *p_attr_rise, *p_attr_size, *p_attr_family;
+	PangoAttribute *p_attr_weight, *p_attr_style;
 
 	p_attr_size = pango_attr_size_new ((int) (gp_cairo_enhanced_fontsize*PANGO_SCALE));
 	p_attr_size->start_index = start;
@@ -1139,6 +1212,16 @@ void gp_cairo_add_attr(plot_struct *plot, PangoAttrList * AttrList, int start, i
 	p_attr_family->start_index = start;
 	p_attr_family->end_index = end;
 	pango_attr_list_insert (AttrList, p_attr_family);
+
+	p_attr_weight = pango_attr_weight_new (plot->fontweight);
+	p_attr_weight->start_index = start;
+	p_attr_weight->end_index = end;
+	pango_attr_list_insert (AttrList, p_attr_weight);
+
+	p_attr_style = pango_attr_style_new (plot->fontstyle);
+	p_attr_style->start_index = start;
+	p_attr_style->end_index = end;
+	pango_attr_list_insert (AttrList, p_attr_style);
 }
 
 /* add a blank character to the text string with a custom shape */
@@ -1192,14 +1275,16 @@ void gp_cairo_enhanced_flush(plot_struct *plot)
 	if (!gp_cairo_enhanced_opened_string)
 		return;
 
-	FPRINTF((stderr, "enhanced flush str=\"%s\" font=%s op=%d sf=%d wf=%d base=%f os=%d\n",
+	FPRINTF((stderr, "enhanced flush str=\"%s\" font=%s op=%d sf=%d wf=%d base=%f os=%d wt=%d sl=%d\n",
 		enhanced_text,
 		gp_cairo_enhanced_font,
 		gp_cairo_enhanced_overprint,
 		gp_cairo_enhanced_showflag,
 		gp_cairo_enhanced_widthflag,
 		gp_cairo_enhanced_base,
-		gp_cairo_enhanced_opened_string ));
+		gp_cairo_enhanced_opened_string,
+		plot->fontweight,
+		plot->fontstyle ));
 
 	gp_cairo_enhanced_opened_string = FALSE;
 
@@ -1244,6 +1329,8 @@ void gp_cairo_enhanced_flush(plot_struct *plot)
 		pango_attr_list_unref( gp_cairo_enhanced_save_AttrList );
 		/* invert the size, so we will go back to the saved state */
 		save_logical_rect.width = -save_logical_rect.width;
+		/* EAM FIXME:  Zero height necessary but I don't understand why */
+		save_logical_rect.height = 0;
 		/* adding a blank character with the corresponding shape */
 		gp_cairo_add_shape(save_logical_rect,start);
 
@@ -1276,9 +1363,8 @@ void gp_cairo_enhanced_flush(plot_struct *plot)
 		current_desc = pango_font_description_new ();
 		pango_font_description_set_family (current_desc, gp_cairo_enhanced_get_fontname(plot));
 		pango_font_description_set_size(current_desc,(int) gp_cairo_enhanced_fontsize*PANGO_SCALE);
-		pango_font_description_set_weight (current_desc, 
-			plot->fontweight ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL);
-		pango_font_description_set_style (current_desc, 
+		pango_font_description_set_weight (current_desc, plot->fontweight);
+		pango_font_description_set_style (current_desc,
 			plot->fontstyle ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL);
 
 		pango_layout_set_font_description (current_layout, current_desc);
@@ -1286,7 +1372,7 @@ void gp_cairo_enhanced_flush(plot_struct *plot)
 		pango_layout_get_extents(current_layout, &current_ink_rect, &current_logical_rect);
 		g_object_unref (current_layout);
 
-		/* calculate the distance to remove to center the overprinted text */ 
+		/* calculate the distance to remove to center the overprinted text */
 		underprinted_logical_rect.width = -(underprinted_logical_rect.width + current_logical_rect.width)/2;
 		overprinted_width = current_logical_rect.width;
 
@@ -1322,9 +1408,8 @@ void gp_cairo_enhanced_flush(plot_struct *plot)
 		hide_desc = pango_font_description_new ();
 		pango_font_description_set_family (hide_desc, gp_cairo_enhanced_get_fontname(plot));
 		pango_font_description_set_size(hide_desc,(int) gp_cairo_enhanced_fontsize*PANGO_SCALE);
-		pango_font_description_set_weight (hide_desc, 
-			plot->fontweight ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL);
-		pango_font_description_set_style (hide_desc, 
+		pango_font_description_set_weight (hide_desc, plot->fontweight);
+		pango_font_description_set_style (hide_desc,
 			plot->fontstyle ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL);
 		pango_layout_set_font_description (hide_layout, hide_desc);
 		pango_font_description_free (hide_desc);
@@ -1346,16 +1431,15 @@ void gp_cairo_enhanced_flush(plot_struct *plot)
 	if (!gp_cairo_enhanced_widthflag) {
 		/* the idea is to use a blank character, drawn with the inverted width of the text*/
 		/* we first compute the size of the text */
-	
+
 		/* Create a PangoLayout, set the font and text */
 		zerowidth_layout = gp_cairo_create_layout (plot->cr);
 		pango_layout_set_text (zerowidth_layout, enhanced_text_utf8, -1);
 		zerowidth_desc = pango_font_description_new ();
 		pango_font_description_set_family (zerowidth_desc, gp_cairo_enhanced_get_fontname(plot));
 		pango_font_description_set_size(zerowidth_desc,(int) gp_cairo_enhanced_fontsize*PANGO_SCALE);
-		pango_font_description_set_weight (zerowidth_desc, 
-			plot->fontweight ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL);
-		pango_font_description_set_style (zerowidth_desc, 
+		pango_font_description_set_weight (zerowidth_desc, plot->fontweight);
+		pango_font_description_set_style (zerowidth_desc,
 			plot->fontstyle ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL);
 		pango_layout_set_font_description (zerowidth_layout, zerowidth_desc);
 		pango_font_description_free (zerowidth_desc);
@@ -1441,9 +1525,16 @@ void gp_cairo_enhanced_open(plot_struct *plot, char* fontname, double fontsize,
 	}
 
 	if (!gp_cairo_enhanced_opened_string) {
+		// Strip off Bold or Italic and apply immediately
+		// Is it really necessary to preserve plot->fontname?
+		char *save_plot_font = strdup(plot->fontname);
+		gp_cairo_set_font(plot, fontname, plot->fontsize);
+		strncpy(gp_cairo_enhanced_font, plot->fontname, sizeof(gp_cairo_enhanced_font));
+		strcpy(plot->fontname, save_plot_font);
+		free(save_plot_font);
+
 		gp_cairo_enhanced_opened_string = TRUE;
 		gp_cairo_enhanced_char = gp_cairo_enhanced_string;
-		strncpy(gp_cairo_enhanced_font, fontname, sizeof(gp_cairo_enhanced_font));
 		gp_cairo_enhanced_fontsize = fontsize*plot->oversampling_scale;
 		gp_cairo_enhanced_base = base*plot->oversampling_scale;
 		gp_cairo_enhanced_showflag = showflag;
@@ -1465,7 +1556,7 @@ void gp_cairo_enhanced_init(plot_struct *plot, int len)
 	/* also draw any open polygon set */
 	gp_cairo_end_polygon(plot);
 
-	
+
 	gp_cairo_enhanced_string = (char*) malloc(len+1);
 	gp_cairo_enhanced_opened_string = FALSE;
 	gp_cairo_enhanced_overprint = FALSE;
@@ -1489,10 +1580,10 @@ void gp_cairo_enhanced_finish(plot_struct *plot, int x, int y)
 	pango_layout_set_attributes (layout, gp_cairo_enhanced_AttrList);
 
 	pango_layout_get_extents(layout, &ink_rect, &logical_rect);
-
+	
 	/* NB: See explanatory comments in gp_cairo_draw_text() */
 	vert_just = 0.8 * (float)(plot->fontsize * plot->oversampling_scale);
-	
+
 	arg = plot->text_angle * M_PI/180;
 	enh_x = x - vert_just * sin(arg);
 	enh_y = y - vert_just * cos(arg);
@@ -1520,7 +1611,8 @@ void gp_cairo_enhanced_finish(plot_struct *plot, int x, int y)
 	/* angle in radians */
 	cairo_rotate(plot->cr, -arg);
 
-	cairo_set_source_rgb(plot->cr, plot->color.r, plot->color.g, plot->color.b);
+	cairo_set_source_rgba(plot->cr, plot->color.r, plot->color.g, plot->color.b,
+				1. - plot->color.alpha);
 	/* Inform Pango to re-layout the text with the new transformation */
 	pango_cairo_update_layout (plot->cr, layout);
 	pango_cairo_show_layout (plot->cr, layout);
@@ -1528,6 +1620,27 @@ void gp_cairo_enhanced_finish(plot_struct *plot, int x, int y)
 	 * Do it by ourselves, or we can get spurious lines on future calls. */
 	cairo_new_path(plot->cr);
 
+#ifdef EAM_BOXED_TEXT
+	/* Update bounding box for boxed label text */
+	pango_layout_get_pixel_extents (layout, &ink_rect, &logical_rect);
+
+	/* Auto-initialization */
+	if (bounding_box[0] < 0 && bounding_box[1] < 0) {
+	    bounding_box[0] = bounding_box[2] = x;
+	    bounding_box[1] = bounding_box[3] = y;
+	}
+
+	/* Would it look better to use logical bounds rather than ink_rect? */
+	if (bounding_box[0] > enh_x + ink_rect.x)
+	    bounding_box[0] = enh_x + ink_rect.x;
+	if (bounding_box[2] < enh_x + ink_rect.x + ink_rect.width)
+	    bounding_box[2] = enh_x + ink_rect.x + ink_rect.width;
+	if (bounding_box[1] > enh_y + ink_rect.y)
+	    bounding_box[1] = enh_y + ink_rect.y;
+	if (bounding_box[3] < enh_y + ink_rect.y + ink_rect.height)
+	    bounding_box[3] = enh_y + ink_rect.y + ink_rect.height;
+#endif
+	
 	/* free the layout object */
 	pango_attr_list_unref( gp_cairo_enhanced_AttrList );
 	g_object_unref (layout);
@@ -1543,32 +1656,37 @@ void gp_cairo_fill(plot_struct *plot, int fillstyle, int fillpar)
 	double red = 0, green = 0, blue = 0, fact = 0;
 
 	switch (fillstyle) {
+	case FS_SOLID: /* solid fill */
+		if (plot->color.alpha > 0) {
+			fillpar = 100. * (1. - plot->color.alpha);
+			/* Fall through to FS_TRANSPARENT_SOLID */
+		} else if (fillpar==100)
+			/* treated as a special case to accelerate common situation */ {
+			cairo_set_source_rgb(plot->cr, plot->color.r, plot->color.g, plot->color.b);
+			FPRINTF((stderr,"solid %lf %lf %lf\n",plot->color.r, plot->color.g, plot->color.b));
+			return;
+		} else {
+			red   = plot->color.r;
+			green = plot->color.g;
+			blue  = plot->color.b;
+
+			fact = (double)(100 - fillpar) /100;
+
+			if (fact >= 0 && fact <= 1) {
+				red   += (1 - red) * fact;
+				green += (1 - green) * fact;
+				blue  += (1 - blue) * fact;
+			}
+			cairo_set_source_rgb(plot->cr, red, green, blue);
+			FPRINTF((stderr,"transparent solid %lf %lf %lf\n",red, green, blue));
+			return;
+		}
 	case FS_TRANSPARENT_SOLID:
 		red   = plot->color.r;
 		green = plot->color.g;
 		blue  = plot->color.b;
 		cairo_set_source_rgba(plot->cr, red, green, blue, (double)fillpar/100.);
 		return;
-	case FS_SOLID: /* solid fill */
-		if (fillpar==100) /* treated as a special case to accelerate common situation */ {
-			cairo_set_source_rgb(plot->cr, plot->color.r, plot->color.g, plot->color.b);
-			FPRINTF((stderr,"solid %lf %lf %lf\n",plot->color.r, plot->color.g, plot->color.b));
-			return;
-		}
-		red   = plot->color.r;
-		green = plot->color.g;
-		blue  = plot->color.b;
-
-		fact = (double)(100 - fillpar) /100;
-		
-		if (fact >= 0 && fact <= 1) {
-			red   += (1 - red) * fact;
-			green += (1 - green) * fact;
-			blue  += (1 - blue) * fact;
-		}
-		cairo_set_source_rgb(plot->cr, red, green, blue);
-		FPRINTF((stderr,"transparent solid %lf %lf %lf\n",red, green, blue));
-		return;
 	case FS_PATTERN: /* pattern fill */
 	case FS_TRANSPARENT_PATTERN:
 		gp_cairo_fill_pattern(plot, fillstyle, fillpar);
@@ -1585,6 +1703,64 @@ void gp_cairo_fill(plot_struct *plot, int fillstyle, int fillpar)
 	}
 }
 
+#ifdef EAM_BOXED_TEXT
+void gp_cairo_boxed_text(plot_struct *plot, int x, int y, int option)
+{
+	int dx, dy;
+
+	switch (option) {
+	case TEXTBOX_INIT:
+		/* Initialize bounding box for this text string */
+		bounding_box[0] = bounding_box[2] = x;
+		bounding_box[1] = bounding_box[3] = y;
+		break;
+
+	case TEXTBOX_OUTLINE:
+		/* Stroke the outline of the bounding box for previous text */
+	case TEXTBOX_BACKGROUNDFILL:
+	case TEXTBOX_GREY:
+		/* Fill the bounding box with background color */
+		/* begin by stroking any open path */
+		gp_cairo_stroke(plot);
+		gp_cairo_end_polygon(plot);
+
+		cairo_save(plot->cr);
+		dx = 0.5 * bounding_xmargin * (float)(plot->fontsize * plot->oversampling_scale);
+		dy = 0.5 * bounding_ymargin * (float)(plot->fontsize * plot->oversampling_scale);
+		if (option == TEXTBOX_GREY)
+		    dy = 0;
+		gp_cairo_move(plot,   bounding_box[0]-dx, bounding_box[1]-dy); 
+		gp_cairo_vector(plot, bounding_box[0]-dx, bounding_box[3]+dy); 
+		gp_cairo_vector(plot, bounding_box[2]+dx, bounding_box[3]+dy); 
+		gp_cairo_vector(plot, bounding_box[2]+dx, bounding_box[1]-dy); 
+		gp_cairo_vector(plot, bounding_box[0]+dx, bounding_box[1]-dy); 
+		cairo_close_path(plot->cr);
+		if (option == TEXTBOX_BACKGROUNDFILL) {
+		    rgb_color *background = &gp_cairo_colorlist[0];
+		    cairo_set_source_rgb(plot->cr, background->r, background->g, background->b);
+		    cairo_fill(plot->cr);
+		} else if (option == TEXTBOX_GREY) {
+		    cairo_set_source_rgba(plot->cr, 0.75, 0.75, 0.75, 0.50);
+		    cairo_fill(plot->cr);
+		} else {
+		    cairo_set_line_width(plot->cr, 1.0*plot->oversampling_scale);
+		    cairo_set_source_rgb(plot->cr,
+			plot->color.r, plot->color.g, plot->color.b);
+		    cairo_stroke(plot->cr);
+		}
+		cairo_restore(plot->cr);
+		break;
+
+	case TEXTBOX_MARGINS: /* Change the margin between text and box */
+		bounding_xmargin = (double)x/100.;
+		bounding_ymargin = (double)y/100.;
+		break;
+
+	default:
+		break;
+	}
+}
+#endif
 
 #define PATTERN_SIZE 8
 
@@ -1614,7 +1790,8 @@ void gp_cairo_fill_pattern(plot_struct *plot, int fillstyle, int fillpar)
 	    cairo_set_source_rgb( pattern_cr, 1.0, 1.0, 1.0);
 
 	cairo_paint(pattern_cr);
-	if (!strcmp(term->name,"pdfcairo")) /* Work-around for poor scaling in cairo <= 1.10 */
+
+	if (!strcmp(term->name,"pdfcairo")) /* Work-around for poor scaling in cairo */
 	    cairo_set_line_width(pattern_cr, PATTERN_SIZE/150.);
 	else
 	    cairo_set_line_width(pattern_cr, PATTERN_SIZE/50.);
@@ -1706,9 +1883,8 @@ void gp_cairo_set_termvar(plot_struct *plot, unsigned int *v_char,
 	desc = pango_font_description_new ();
 	pango_font_description_set_family (desc, plot->fontname);
 	pango_font_description_set_size(desc,(int) (plot->fontsize*PANGO_SCALE*plot->oversampling_scale));
-	pango_font_description_set_weight (desc, 
-		plot->fontweight ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL);
-	pango_font_description_set_style (desc, 
+	pango_font_description_set_weight (desc, plot->fontweight);
+	pango_font_description_set_style (desc,
 		plot->fontstyle ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL);
 	pango_layout_set_font_description (layout, desc);
 	pango_font_description_free (desc);
@@ -1736,9 +1912,9 @@ void gp_cairo_set_termvar(plot_struct *plot, unsigned int *v_char,
 void gp_cairo_solid_background(plot_struct *plot)
 {
 	if (cairo_status (plot->cr)) {
-		printf("Cairo is unhappy: %s\n",
+		fprintf(stderr, "Cairo is unhappy: %s\n",
 			cairo_status_to_string (cairo_status (plot->cr)));
-		exit(0);
+		gp_exit(EXIT_FAILURE);
 	}
 	cairo_set_source_rgb(plot->cr, plot->background.r, plot->background.g, plot->background.b);
 	cairo_paint(plot->cr);
@@ -1747,9 +1923,9 @@ void gp_cairo_solid_background(plot_struct *plot)
 void gp_cairo_clear_background(plot_struct *plot)
 {
 	if (cairo_status (plot->cr)) {
-		printf("Cairo is unhappy: %s\n",
+		fprintf(stderr, "Cairo is unhappy: %s\n",
 			cairo_status_to_string (cairo_status (plot->cr)));
-		exit(0);
+		gp_exit(EXIT_FAILURE);
 	}
 	cairo_set_source_rgba(plot->cr, 0.0, 0.0, 0.0, 0.0);
 	cairo_paint(plot->cr);
@@ -1771,7 +1947,7 @@ const char* gp_cairo_enhanced_get_fontname(plot_struct *plot)
 }
 
 /* BM: New function to determine the default font.
- * On Windows, the "Sans" alias normally is equivalent to 
+ * On Windows, the "Sans" alias normally is equivalent to
  * "Tahoma" but the resolution fails on some systems. */
 const char *
 gp_cairo_default_font(void)
@@ -1832,6 +2008,7 @@ const char* gp_cairo_get_encoding(plot_struct *plot)
 	case S_ENC_CP850 :  return "cp850";
 	case S_ENC_CP852 :  return "cp852";
 	case S_ENC_CP1250 : return "windows-1250";
+	case S_ENC_CP1252 : return "windows-1252";
 	case S_ENC_KOI8_R : return "KOI8-R";
 	case S_ENC_KOI8_U :  return "KOI8-U";
 	case S_ENC_ISO8859_1 : return "ISO-8859-1";
@@ -1978,7 +2155,7 @@ gchar* gp_cairo_convert_symbol_to_unicode(plot_struct *plot, const char* string)
 		SYMB_UNICODE(0x7C,0x007C); /* VERTICAL LINE */
 		SYMB_UNICODE(0x7D,0x007D); /* RIGHT CURLY BRACKET */
 		SYMB_UNICODE(0x7E,0x223C); /* TILDE OPERATOR */
-		
+
 		SYMB_UNICODE(0xA0,0x20AC); /* EURO SIGN */
 		SYMB_UNICODE(0xA1,0x03D2); /* GREEK UPSILON WITH HOOK SYMBOL */
 		SYMB_UNICODE(0xA2,0x2032); /* PRIME minute */
@@ -2071,7 +2248,7 @@ gchar* gp_cairo_convert_symbol_to_unicode(plot_struct *plot, const char* string)
 		SYMB_UNICODE(0xFC,0x23AB); /* RIGHT CURLY BRACKET UPPER HOOK */
 		SYMB_UNICODE(0xFD,0x23AC); /* RIGHT CURLY BRACKET MIDDLE PIECE */
 		SYMB_UNICODE(0xFE,0x23AD); /* RIGHT CURLY BRACKET LOWER HOOK */
-		
+
 		/* to be treated specifically : composed characters */
 		case 0xE2 : /* REGISTERED SIGN, alternate: sans serif */
 			g_unichar_to_utf8(0x00AE,iter_mod);
